#include "boost/lexical_cast.hpp"
#include "ParallelDomainVec.h"
#include "ProblemVec.h"
#include "ProblemInstat.h"
#include "SystemVector.h"
#include "Serializer.h"
#include "BoundaryManager.h"

namespace AMDiS {

  using boost::lexical_cast;

  ParallelDomainVec::ParallelDomainVec(ProblemVec *problem,
				       ProblemInstatVec *problemInstat)
    : ParallelDomainBase(problem, 
			 problemInstat, 
			 problem->getFESpace(0),
			 problem->getRefinementManager(0)),
      probVec(problem)
  {
    FUNCNAME("ParallelDomainVec::ParallelDomainVec()");

    info = probVec->getInfo();
    nComponents = probVec->getNumComponents();

    // Test if all fe spaces are equal. Yet, different fe spaces are not supported for
    // domain parallelization.
    const FiniteElemSpace *fe = probVec->getFESpace(0);
    for (int i = 0; i < nComponents; i++)
      TEST_EXIT(fe == probVec->getFESpace(i))
	("Parallelization does not supported different FE spaces!\n");

    // Create parallel serialization file writer, if needed.
    int writeSerialization = 0;
    GET_PARAMETER(0, name + "->output->write serialization", "%d", &writeSerialization);
    if (writeSerialization)
      problem->getFileWriterList().push_back(new Serializer<ParallelDomainVec>(this));

    int readSerialization = 0;
    GET_PARAMETER(0, name + "->input->read serialization", "%d", &readSerialization);
    if (readSerialization) {
      std::string filename = "";
      GET_PARAMETER(0, name + "->input->serialization filename", &filename);
      filename += ".p" + lexical_cast<std::string>(mpiRank);
      MSG("Start serialization with %s\n", filename.c_str());
      std::ifstream in(filename.c_str());
      deserialize(in);
      in.close();
    }
  }


  void ParallelDomainVec::initParallelization(AdaptInfo *adaptInfo)
  {
    FUNCNAME("ParallelDomainVec::initParallelization()");

    ParallelDomainBase::initParallelization(adaptInfo);

    for (int i = 0; i < nComponents; i++) {
      for (int j = 0; j < nComponents; j++)
 	if (probVec->getSystemMatrix(i, j))
 	  probVec->getSystemMatrix(i, j)->setRankDofs(isRankDof);      

      TEST_EXIT_DBG(probVec->getRHS()->getDOFVector(i))("No rhs vector!\n");
      TEST_EXIT_DBG(probVec->getSolution()->getDOFVector(i))("No solution vector!\n");

      probVec->getRHS()->getDOFVector(i)->setRankDofs(isRankDof);
      probVec->getSolution()->getDOFVector(i)->setRankDofs(isRankDof);
    }

    // === Remove periodic boundary conditions in sequential problem definition. ===

    // Remove periodic boundaries in boundary manager on matrices and vectors.
    for (int i = 0; i < nComponents; i++) {
      for (int j = 0; j < nComponents; j++) {
	DOFMatrix* mat = probVec->getSystemMatrix(i, j);
 	if (mat && mat->getBoundaryManager())
	  removeBoundaryCondition(const_cast<BoundaryIndexMap&>(mat->getBoundaryManager()->getBoundaryConditionMap()));
      }

      if (probVec->getSolution()->getDOFVector(i)->getBoundaryManager())
	removeBoundaryCondition(const_cast<BoundaryIndexMap&>(probVec->getSolution()->getDOFVector(i)->getBoundaryManager()->getBoundaryConditionMap()));

      if (probVec->getRHS()->getDOFVector(i)->getBoundaryManager())
	removeBoundaryCondition(const_cast<BoundaryIndexMap&>(probVec->getRHS()->getDOFVector(i)->getBoundaryManager()->getBoundaryConditionMap()));
    }

    // Remove periodic boundaries on elements in mesh.
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh,  -1, Mesh::CALL_EVERY_EL_PREORDER);
    while (elInfo) {
      elInfo->getElement()->deleteElementData(PERIODIC);
      elInfo = stack.traverseNext(elInfo);
    }    
  }

  void ParallelDomainVec::exitParallelization(AdaptInfo *adaptInfo)
  {
    ParallelDomainBase::exitParallelization(adaptInfo);
  }

  void ParallelDomainVec::solveInitialProblem(AdaptInfo *adaptInfo)
  {   
    ParallelDomainBase::solveInitialProblem(adaptInfo);
    
    synchVectors(*(probVec->getSolution()));
  }

  void ParallelDomainVec::solve()
  {
    FUNCNAME("ParallelDomainVec::solve()");

#ifdef _OPENMP
    double wtime = omp_get_wtime();
#endif
    clock_t first = clock();

    fillPetscMatrix(probVec->getSystemMatrix(), probVec->getRHS());      
    solvePetscMatrix(*(probVec->getSolution()));

#ifdef _OPENMP
    INFO(info, 8)("solution of discrete system needed %.5f seconds system time / %.5f seconds wallclock time\n",
		   TIME_USED(first, clock()),
		   omp_get_wtime() - wtime);
#else
    INFO(info, 8)("solution of discrete system needed %.5f seconds\n",
		   TIME_USED(first, clock()));
#endif    
  }


  void ParallelDomainVec::removeBoundaryCondition(BoundaryIndexMap& boundaryMap)
  {
    BoundaryIndexMap::iterator it = boundaryMap.begin();
    while (it != boundaryMap.end()) {
      if (it->second->isPeriodic())
	boundaryMap.erase(it++);
      else
	++it;      
    }    
  }

}
